import os
from glob import glob
from os.path import join, basename
from textwrap import dedent

from ..util import replace_ext, pairwise, gdal, osr


def browse_georeference(source_dir: os.PathLike, target_dir_data: os.PathLike, target_dir_meta: os.PathLike, browse_report: dict, browse: dict):
    for filename in glob(join(source_dir, '*')):
        target_filename_data = join(target_dir_data, replace_ext(basename(filename), '.tif'))
        target_filename_meta = join(target_dir_meta, replace_ext(basename(filename), '.xml'))
        apply_browse_report_georeference(filename, target_filename_data, browse)

        generate_gsc(filename, target_filename_meta, browse_report, browse)


def apply_browse_report_georeference(input_filename: os.PathLike, target_filename: os.PathLike, browse: dict):
    ds = gdal.GetDriverByName('GTiff').CreateCopy(target_filename, gdal.Open(input_filename))
    type_ = browse['type']

    if type_ == 'rectified_browse':
        size_x, size_y = ds.RasterXSize, ds.RasterYSize
        low, high = browse['rectified']['coord_list']

        minx, miny = low
        maxx, maxy = high

        ds.SetGeoTransform([
            minx, (maxx - minx) / size_x, 0,
            maxy, 0, (miny - maxy) / size_y,
        ])

    elif type_ == 'footprint_browse':
        col_rows = browse['footprint']['col_row_list']
        coords = browse['footprint']['coord_list']

        if col_rows[0] == col_rows[-1] and coords[0] == coords[-1]:
            col_rows = col_rows[:-1]
            coords = coords[:-1]

        gcps = [
            gdal.GCP(coord[1], coord[0], 0, col_row[0], col_row[1])
            for col_row, coord in zip(col_rows, coords)
        ]

        sr = osr.SpatialReference()
        sr.ImportFromEPSG(4326)
        ds.SetGCPs(gcps, sr)

    elif type_ == 'model_in_geotiff_browse':
        # nothing to do in this case
        pass

    elif type_ == 'regular_grid_browse':
        col_node_number = browse['regular_grid']['col_node_number']
        row_node_number = browse['regular_grid']['row_node_number']
        col_step = browse['regular_grid']['col_step']
        row_step = browse['regular_grid']['row_step']

        coord_lists = browse['regular_grid']['coord_lists']

        range_x = frange(0.0, row_node_number * row_step, row_step)
        range_y = frange(0.0, col_node_number * col_step, col_step)
        pixels = [(x, y) for x in range_x for y in range_y]

        coords = [
            coord
            for coord in coord_list for coord_list in coord_lists
        ]

        gcps = [
            gdal.GCP(coord[1], coord[0], 0, col_row[0], col_row[1])
            for col_row, coord in zip(col_rows, coords)
        ]

        sr = osr.SpatialReference()
        sr.ImportFromEPSG(4326)
        ds.SetGCPs(gcps, sr)

    else:
        raise Exception

    del ds


# copied from: https://pynative.com/python-range-for-float-numbers/
def frange(start, stop=None, step=None):
    # if stop and step argument is None set start=0.0 and step = 1.0
    start = float(start)
    if stop == None:
        stop = start + 0.0
        start = 0.0
    if step == None:
        step = 1.0

    count = 0
    while True:
        temp = float(start + count * step)
        if step > 0 and temp >= stop:
            break
        elif step < 0 and temp <= stop:
            break
        yield temp

        count += 1


def generate_gsc(input_filename: os.PathLike, target_filename: os.PathLike, browse_report: dict, browse: dict):
    footprint = ''

    type_ = browse['type']
    if type_ == 'rectified_browse':
        low, high = browse['rectified']['coord_list']
        footprint = ' '.join(str(v) for v in [
            low[0], low[1],
            high[0], low[1],
            high[0], high[1],
            low[0], high[1],
            low[0], low[1],
        ])

    elif type_ == 'footprint_browse':
        footprint = ' '.join([
            f'{y} {x}'
            for y, x in browse['footprint']['coord_list']
        ])

    elif type_ == 'model_in_geotiff_browse':
        ds = gdal.Open(input_filename)
        gt = ds.GetGeoTransform()

        low = (gt[3], gt[0])
        high = (gt[3] + ds.RasterYSize * gt[5], gt[0] + ds.RasterXSize * gt[1])

        footprint = ' '.join(str(v) for v in [
            low[0], low[1],
            high[0], low[1],
            high[0], high[1],
            low[0], high[1],
            low[0], low[1],
        ])

    elif type_ == 'regular_grid_browse':
        coord_lists = browse['regular_grid']['coord_lists']
        coords = coord_lists[0] + [
            coord_list[-1]
            for coord_list in coord_lists[1:-1]
        ] + reversed(coord_lists[-1]) + [
            coord_list[0]
            for coord_list in coord_lists[-1::-1]
        ]

        footprint = ' '.join([
            f'{y} {x}'
            for y, x in coords
        ])

    xmlcontent = dedent(f"""\
        <?xml version='1.0' encoding='UTF-8'?>
        <gsc:report
            xmlns:sar="http://earth.esa.int/sar"
            xmlns:gml="http://www.opengis.net/gml"
            xmlns:eop="http://earth.esa.int/eop"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xmlns:opt="http://earth.esa.int/opt"
            xmlns:gsc="http://earth.esa.int/gsc"
            xmlns:atm="http://earth.esa.int/atm"
            xmlns:xlink="http://www.w3.org/1999/xlink" version="2.0">
          <gsc:responsibleOrgName>{browse_report['responsible_org_name']}</gsc:responsibleOrgName>
          <gsc:reportType>BROWSE</gsc:reportType>
          <gsc:dateTime>{browse_report['date_time']}</gsc:dateTime>
          <gsc:orderReference></gsc:orderReference>
          <gsc:opt_metadata version="1.2.1">
            <gml:metaDataProperty>
              <gsc:EarthObservationMetaData>
                <eop:identifier>{browse['browse_identifier']}</eop:identifier>
              </gsc:EarthObservationMetaData>
              </gml:metaDataProperty>
              <gml:validTime>
              <gml:TimePeriod>
                <gml:beginPosition>{browse['start_time']}</gml:beginPosition>
                <gml:endPosition>{browse['end_time']}</gml:endPosition>
              </gml:TimePeriod>
              </gml:validTime>
              <gml:using>
                <eop:EarthObservationEquipment>
                </eop:EarthObservationEquipment>
              </gml:using>
              <gml:target>
                <eop:Footprint>
                  <gml:multiExtentOf>
                    <gml:MultiSurface srsName="EPSG:4326">
                      <gml:surfaceMembers>
                        <gml:Polygon>
                          <gml:exterior>
                            <gml:LinearRing>
                              <gml:posList>{footprint}</gml:posList>
                            </gml:LinearRing>
                          </gml:exterior>
                        </gml:Polygon>
                      </gml:surfaceMembers>
                    </gml:MultiSurface>
                  </gml:multiExtentOf>
                </eop:Footprint>
              </gml:target>
            <gml:resultOf/>
          </gsc:opt_metadata>
        </gsc:report>
    """)
    with open(target_filename, 'w') as ff:
        ff.write(xmlcontent)
    return xmlcontent
